from typing import Callable, Dict, Any

from slack_sdk.signature import SignatureVerifier

from slack_bolt.logger import get_bolt_logger
from .middleware import Middleware
from slack_bolt.request import BoltRequest
from slack_bolt.response import BoltResponse


class RequestVerification(Middleware):  # type: ignore
    def __init__(self, signing_secret: str):
        self.verifier = SignatureVerifier(signing_secret=signing_secret)
        self.logger = get_bolt_logger(RequestVerification)

    def process(
        self, *, req: BoltRequest, resp: BoltResponse, next: Callable[[], BoltResponse],
    ) -> BoltResponse:
        if self._can_skip(req.payload):
            return next()

        body = req.body
        timestamp = req.headers.get("x-slack-request-timestamp", ["0"])[0]
        signature = req.headers.get("x-slack-signature", [""])[0]
        if self.verifier.is_valid(body, timestamp, signature):
            return next()
        else:
            self._debug_log_error(signature, timestamp, body)
            return self._build_error_response()

    # -----------------------------------------

    @staticmethod
    def _can_skip(payload: Dict[str, Any]) -> bool:
        return payload is not None and payload.get("ssl_check", None) == "1"

    @staticmethod
    def _build_error_response() -> BoltResponse:
        return BoltResponse(status=401, body={"error": "invalid request"})

    def _debug_log_error(self, signature, timestamp, body) -> None:
        self.logger.info(
            "Invalid request signature detected "
            f"(signature: {signature}, timestamp: {timestamp}, body: {body})"
        )



import inspect
from logging import Logger
from typing import Callable, List

from slack_bolt.kwargs_injection import build_required_kwargs
from slack_bolt.listener_matcher import ListenerMatcher
from slack_bolt.request import BoltRequest
from slack_bolt.response import BoltResponse
from .listener import Listener
from ..logger import get_bolt_app_logger
from ..middleware import Middleware


class CustomListener(Listener):
    app_name: str
    func: Callable[..., BoltResponse]
    matchers: List[ListenerMatcher]
    middleware: List[Middleware]  # type: ignore
    auto_acknowledgement: bool
    arg_names: List[str]
    logger: Logger

    def __init__(
        self,
        *,
        app_name: str,
        func: Callable[..., BoltResponse],
        matchers: List[ListenerMatcher],
        middleware: List[Middleware],  # type: ignore
        auto_acknowledgement: bool = False,
    ):
        self.app_name = app_name
        self.func = func
        self.matchers = matchers
        self.middleware = middleware
        self.auto_acknowledgement = auto_acknowledgement
        self.arg_names = inspect.getfullargspec(func).args
        self.logger = get_bolt_app_logger(app_name, self.func)

    def run_ack_function(
        self, *, request: BoltRequest, response: BoltResponse,
    ) -> BoltResponse:
        return self.func(
            **build_required_kwargs(
                logger=self.logger,
                required_arg_names=self.arg_names,
                request=request,
                response=response,
            )
        )



from typing import Optional, Union, Dict, List

from slack_sdk import WebClient
from slack_sdk.models import JsonObject

from slack_bolt.error import BoltError
from slack_bolt.version import __version__ as bolt_version


def create_web_client(token: Optional[str] = None) -> WebClient:
    return WebClient(token=token, user_agent_prefix=f"Bolt/{bolt_version}",)


def convert_to_dict_list(objects: List[Union[Dict, JsonObject]]) -> List[Dict]:
    return [_to_dict(elm) for elm in objects]


def _to_dict(obj: Union[Dict, JsonObject]) -> Dict:
    if isinstance(obj, dict):
        return obj
    if isinstance(obj, JsonObject) or hasattr(obj, "to_dict"):
        return obj.to_dict()
    raise BoltError(f"{obj} (type: {type(obj)}) is unsupported")



from typing import Dict, Optional, List, Union, Any

from slack_bolt.context.context import BoltContext
from slack_bolt.request.internals import (
    parse_query,
    parse_payload,
    build_normalized_headers,
    build_context,
    extract_content_type,
)


class BoltRequest:
    body: str
    query: Dict[str, List[str]]
    headers: Dict[str, List[str]]
    content_type: Optional[str]
    payload: Dict[str, Any]
    context: BoltContext

    def __init__(
        self,
        *,
        body: str,
        query: Optional[Union[str, Dict[str, str], Dict[str, List[str]]]] = None,
        # many framework use Dict[str, str] but the reality is Dict[str, List[str]]
        headers: Optional[Dict[str, Union[str, List[str]]]] = None,
        context: Optional[Dict[str, str]] = None,
    ):
        self.body = body
        self.query = parse_query(query)
        self.headers = build_normalized_headers(headers)
        self.content_type = extract_content_type(self.headers)
        self.payload = parse_payload(self.body, self.content_type)
        self.context = build_context(
            BoltContext(context if context else {}), self.payload
        )
